using UnityEngine;
using UnityEngine.UI;
using System;
using System.Text;
using System.Text.RegularExpressions;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// **********************************
// VAMLightsabers 1.0
// **********************************

namespace VAMLightsabersPlugin
{
    public partial class VAMLightsabers : MVRScript
    {	
	    // Static string
        public static string PLUGIN_PATH;
        private static string SABER_BUNDLE_PATH = "/assets/vamlightsabers_assets.assetbundle";
        private static string HILTS_PATH = "assets/vamlightsabers/materials/hilts/"; // whatever, just one of the folders with all the hilts names
        private static string HILTS_PREFABS_PATH = "assets/vamlightsabers/prefabs/hilts/";
        private static string ON_SOUNDS_PATH = "assets/vamlightsabers/sounds/on/"; 
        private static string OFF_SOUNDS_PATH = "assets/vamlightsabers/sounds/off/"; 
        private static string LOOPS_SOUNDS_PATH = "assets/vamlightsabers/sounds/loops/"; 
        
        public static string INFO_PREFIX = "VAMLightsabers: ";
        public static string EXCEPTION_PREFIX = "VAMLightsabers - Exception caught: ";

        public static string DEFAULT_SKIN = "deku";
        
        // JSONStorables
        protected JSONStorableColor SaberOpt_ColorBladeA;
        protected JSONStorableColor SaberOpt_ColorBladeGlowA;
        protected JSONStorableColor SaberOpt_ColorLightningA;
        protected JSONStorableColor SaberOpt_ColorTendrilsA;
        protected JSONStorableColor SaberOpt_ColorEmitterA;
        protected JSONStorableColor SaberOpt_ColorSwitchA;
        protected JSONStorableColor SaberOpt_ColorHandleA;
        protected JSONStorableColor SaberOpt_ColorPowerCellA;
        
        protected JSONStorableColor SaberOpt_ColorBladeB;
        protected JSONStorableColor SaberOpt_ColorBladeGlowB;
        protected JSONStorableColor SaberOpt_ColorLightningB;
        protected JSONStorableColor SaberOpt_ColorTendrilsB;
        protected JSONStorableColor SaberOpt_ColorEmitterB;
        protected JSONStorableColor SaberOpt_ColorSwitchB;
        protected JSONStorableColor SaberOpt_ColorHandleB;
        protected JSONStorableColor SaberOpt_ColorPowerCellB;
        
        protected JSONStorableFloat SaberOpt_BladeAOffset;
        protected JSONStorableStringChooser SaberOpt_BladeA;
        protected JSONStorableBool SaberOpt_BladeAEnableLightning;
        protected JSONStorableBool SaberOpt_BladeAEnableTendrils;
        protected JSONStorableStringChooser SaberOpt_EmitterA;
        protected JSONStorableStringChooser SaberOpt_SwitchA;
        protected JSONStorableStringChooser SaberOpt_HandleA;
        protected JSONStorableStringChooser SaberOpt_PowerCellA;
        protected JSONStorableFloat SaberOpt_SwitchAOffset;
        protected JSONStorableFloat SaberOpt_HandleAOffset;
        protected JSONStorableFloat SaberOpt_PowerCellAOffset;

        protected JSONStorableFloat SaberOpt_BladeBOffset;
        protected JSONStorableStringChooser SaberOpt_BladeB;
        protected JSONStorableBool SaberOpt_BladeBEnableLightning;
        protected JSONStorableBool SaberOpt_BladeBEnableTendrils;
        protected JSONStorableStringChooser SaberOpt_EmitterB;
        protected JSONStorableStringChooser SaberOpt_SwitchB;
        protected JSONStorableStringChooser SaberOpt_HandleB;
        protected JSONStorableStringChooser SaberOpt_PowerCellB;
        protected JSONStorableFloat SaberOpt_SwitchBOffset;
        protected JSONStorableFloat SaberOpt_HandleBOffset;
        protected JSONStorableFloat SaberOpt_PowerCellBOffset;
        
        protected JSONStorableBool SaberOpt_EnableDualBladed;
        protected JSONStorableBool SaberOpt_DualBladedOnlyOneSwitch;
        protected JSONStorableStringChooser SaberOpt_DualBladedCellCount;
        protected JSONStorableFloat SaberOpt_DualBladeHandleOffset;
        protected JSONStorableFloat SaberOpt_DualBladeHandleRotation;
        
        protected JSONStorableBool SaberOpt_StartsOn;
        protected JSONStorableFloat SaberOpt_SwitchSoundsVolume;
        protected JSONStorableFloat SaberOpt_LoopSoundsVolume;
        
        protected JSONStorableStringChooser SaberOpt_OnSound;
        protected JSONStorableBool SaberOpt_OnSoundRandom;
        
        protected JSONStorableStringChooser SaberOpt_OffSound;
        protected JSONStorableBool SaberOpt_OffSoundRandom;
        
        protected JSONStorableStringChooser SaberOpt_LoopSound;
        
        protected JSONStorableFloat SaberOpt_BladesFresnelScale;
        protected JSONStorableFloat SaberOpt_BladesFresnelPower;
        protected JSONStorableFloat SaberOpt_PartsRoughness;
        
		protected List<string> _hiltsChoices = new List<string>();
		protected List<string> _bladesChoices = new List<string>();
		protected List<string> _cellCountChoices = new List<string>() {"0","1","2"};
		
		protected List<string> _onSndChoices = new List<string>();
		protected List<AudioClip> _onSndSounds = new List<AudioClip>();
		
		protected List<string> _offSndChoices = new List<string>();
		protected List<AudioClip> _offSndSounds = new List<AudioClip>();
		
		protected List<string> _loopSndChoices = new List<string>();
		protected List<AudioClip> _loopSndSounds = new List<AudioClip>();
        
		// JSONStorables Actions
		public JSONStorableAction A_LS_SwitchOn;
		public JSONStorableAction A_LS_SwitchOff;
			
		// JSONStorables UI
		protected JSONStorableStringChooser LeftUISelector;
		protected List<string> uisLeftChoices = new List<string> {"Main Saber Colors", "Secondary Saber Colors"};

		protected JSONStorableStringChooser RightUISelector;
		protected List<string> uisChoices = new List<string> {"Main Saber Options", "Secondary Saber Options", "Global Saber Options"};
		
        // The usual help section
        protected JSONStorableString helpText;
        
        // VAMS External values
        public JSONStorableBool VAMLSValue_Ready;
        
        // Global UI components
        protected UIDynamicTextField helpWindow;
        
        // Dynamic UI
        protected List<UIDynamic> mainSaberColorOptsUIs = new List<UIDynamic>();
        protected List<UIDynamic> secSaberColorOptsUIs = new List<UIDynamic>();
        protected List<UIDynamic> mainSaberOptsUIs = new List<UIDynamic>();
        protected List<UIDynamic> secSaberOptsUIs = new List<UIDynamic>();
        protected List<UIDynamic> globalSaberOptsUIs = new List<UIDynamic>();

        // Game Objects, components, assets and assets lists
        protected Dictionary<string, List<GameObject>> _hiltsGameobjects = new Dictionary<string, List<GameObject>>();
        
        protected GameObject VAMLSGO;
        
        protected Transform _TRSaberMain;
        protected Transform _TRSaberParent;
        protected Transform _TRBladeParent;
        protected Transform _TRBlade2Parent;
        protected Transform _TRBladeTypes;
        protected Transform _TRBladeTypes2;
        protected Transform _TRHiltParent;
        
        protected Transform _TRLoopSnd;
        protected Transform _TROnSnd;
        protected Transform _TROffSnd;
        
        protected AudioSource _ASLoopSnd;
        protected AudioSource _ASOnSnd;
        protected AudioSource _ASOffSnd;
        
        protected Animation _ANIMSaber;
        
        // Textures
        private RenderTexture _camRT;
        
        // Dynamic UI
        protected List<UIDynamic> ignoreLayoutUI = new List<UIDynamic>();
        
        // Miscs
        protected float _vamexmTimer = 0f;
        protected bool _vamlsInitialized = false;
        protected bool _isBundleLoaded = false;
        protected float _partBoundsHeight = 0f;
        
        // Assetbundle
        private Request bundleRequest = null;
        
        // Security
        protected bool _isSceneLoading { get { return SuperController.singleton.isLoading; } }
        protected bool _isSceneFrozen { get { return SuperController.singleton.freezeAnimation; } }

        // Some colors
        HSVColor hsvcDefault = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
        HSVColor hsvcDefaultBlade = HSVColorPicker.RGBToHSV(0.3f, 0.98f, 0.35f);
        
		public override void Init()
        {			 
            try
            {
                if (containingAtom.type != "Empty")
                {
                    SuperController.LogError(EXCEPTION_PREFIX + "Please add VAMLive Camera to an Empty atom");
                    return;
                }
                
                PLUGIN_PATH = Utils.GetPluginPath(this);
                
                /* *** Temporary UIDynamic *** */

                UIDynamicTextField tmpTextfield;
                UIDynamicPopup tmpPopup;
                UIDynamicSlider tmpSlider;
                UIDynamicButton tmpButton;
                UIDynamicToggle tmpToggle;
                UIDynamicColorPicker tmpColor;
                UIDynamic tmpSpacer;

                tmpButton = CreateButton("Switch on", false);
                tmpButton.button.onClick.AddListener(() => { OnLightsaberSwitchOn(); } );
                SetupButtonWithLayout(tmpButton, 250f);
                
                tmpButton = CreateButton("Switch off", false);
                tmpButton.button.onClick.AddListener(() => { OnLightsaberSwitchOff(); } );
                SetupButtonWithoutLayout(tmpButton, 250f, new Vector2(260, -35));

                /* ********************************* */
                /* **** LEFT UI                 **** */
                /* ********************************* */
                
                LeftUISelector = new JSONStorableStringChooser("Left UI Selector", uisLeftChoices, uisLeftChoices.First(), "Show") {isStorable=false,isRestorable=false};
                LeftUISelector.setCallbackFunction += (val) => { OnChangeLeftUISelection(); };
                UIDynamicPopup LUISelsp = CreateScrollablePopup(LeftUISelector, false);
                LUISelsp.labelWidth = 125f;

                tmpSpacer = CreateSpacer(false);
                Utils.SetupSpacer(tmpSpacer, 20f);
                
                // **** Main saber colors ****
                tmpTextfield = Utils.CreateStaticDescriptionText(this,"ColorsMain","<color=#000><size=28><b>Main lightsaber colors</b></size></color>\n<size=28>Change the colors for the main blade and VFX (one handed lightsaber).</size>",false,95,TextAnchor.MiddleLeft);
                mainSaberColorOptsUIs.Add((UIDynamic) tmpTextfield);
                
                SaberOpt_ColorBladeA = new JSONStorableColor("SaberOpt_ColorBladeA", hsvcDefaultBlade);
                SaberOpt_ColorBladeA.setJSONCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpColor = CreateColorPicker(SaberOpt_ColorBladeA, false);
                tmpColor.label = "Blade Color";
                mainSaberColorOptsUIs.Add((UIDynamic) tmpColor);
                
                SaberOpt_ColorBladeGlowA = new JSONStorableColor("SaberOpt_ColorBladeGlowA", hsvcDefaultBlade);
                SaberOpt_ColorBladeGlowA.setJSONCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpColor = CreateColorPicker(SaberOpt_ColorBladeGlowA, false);
                tmpColor.label = "Blade Glow Color";
                mainSaberColorOptsUIs.Add((UIDynamic) tmpColor);
                
                SaberOpt_ColorLightningA = new JSONStorableColor("SaberOpt_ColorLightningA", hsvcDefaultBlade);
                SaberOpt_ColorLightningA.setJSONCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpColor = CreateColorPicker(SaberOpt_ColorLightningA, false);
                tmpColor.label = "Lightning Color";
                mainSaberColorOptsUIs.Add((UIDynamic) tmpColor);
                
                SaberOpt_ColorTendrilsA = new JSONStorableColor("SaberOpt_ColorTendrilsA", hsvcDefaultBlade);
                SaberOpt_ColorTendrilsA.setJSONCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpColor = CreateColorPicker(SaberOpt_ColorTendrilsA, false);
                tmpColor.label = "Tendrils Color";
                mainSaberColorOptsUIs.Add((UIDynamic) tmpColor);
                
                tmpSpacer = CreateSpacer(false);
                Utils.SetupSpacer(tmpSpacer, 20f);
                mainSaberColorOptsUIs.Add((UIDynamic) tmpSpacer);
                
                tmpTextfield = Utils.CreateStaticDescriptionText(this,"ColorsParts","<color=#000><size=28><b>Main lightsaber parts</b></size></color>\n<size=28>Change the colors for parts of the main lightsaber.</size>",false,95,TextAnchor.MiddleLeft);
                mainSaberColorOptsUIs.Add((UIDynamic) tmpTextfield);
                
                SaberOpt_ColorEmitterA = new JSONStorableColor("SaberOpt_ColorEmitterA", hsvcDefault);
                SaberOpt_ColorEmitterA.setJSONCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpColor = CreateColorPicker(SaberOpt_ColorEmitterA, false);
                tmpColor.label = "Emitter Color";
                mainSaberColorOptsUIs.Add((UIDynamic) tmpColor);
                
                SaberOpt_ColorSwitchA = new JSONStorableColor("SaberOpt_ColorSwitchA", hsvcDefault);
                SaberOpt_ColorSwitchA.setJSONCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpColor = CreateColorPicker(SaberOpt_ColorSwitchA, false);
                tmpColor.label = "Switch Color";
                mainSaberColorOptsUIs.Add((UIDynamic) tmpColor);
                
                SaberOpt_ColorHandleA = new JSONStorableColor("SaberOpt_ColorHandleA", hsvcDefault);
                SaberOpt_ColorHandleA.setJSONCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpColor = CreateColorPicker(SaberOpt_ColorHandleA, false);
                tmpColor.label = "Handle Color";
                mainSaberColorOptsUIs.Add((UIDynamic) tmpColor);
                
                SaberOpt_ColorPowerCellA = new JSONStorableColor("SaberOpt_ColorPowerCellA", hsvcDefault);
                SaberOpt_ColorPowerCellA.setJSONCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpColor = CreateColorPicker(SaberOpt_ColorPowerCellA, false);
                tmpColor.label = "Power cell Color";
                mainSaberColorOptsUIs.Add((UIDynamic) tmpColor);
                
                // **** Secondary saber colors ****
                tmpTextfield = Utils.CreateStaticDescriptionText(this,"ColorsSec","<color=#000><size=28><b>Secondary lightsaber colors</b></size></color>\n<size=28>Change the colors for the secondary lightsaber.</size>",false,95,TextAnchor.MiddleLeft);
                secSaberColorOptsUIs.Add((UIDynamic) tmpTextfield);

                SaberOpt_ColorBladeB = new JSONStorableColor("SaberOpt_ColorBladeB", hsvcDefaultBlade);
                SaberOpt_ColorBladeB.setJSONCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpColor = CreateColorPicker(SaberOpt_ColorBladeB, false);
                tmpColor.label = "Blade Color";
                secSaberColorOptsUIs.Add((UIDynamic) tmpColor);
                
                SaberOpt_ColorBladeGlowB = new JSONStorableColor("SaberOpt_ColorBladeGlowB", hsvcDefaultBlade);
                SaberOpt_ColorBladeGlowB.setJSONCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpColor = CreateColorPicker(SaberOpt_ColorBladeGlowB, false);
                tmpColor.label = "Blade Glow Color";
                secSaberColorOptsUIs.Add((UIDynamic) tmpColor);
                
                SaberOpt_ColorLightningB = new JSONStorableColor("SaberOpt_ColorLightningB", hsvcDefaultBlade);
                SaberOpt_ColorLightningB.setJSONCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpColor = CreateColorPicker(SaberOpt_ColorLightningB, false);
                tmpColor.label = "Lightning Color";
                secSaberColorOptsUIs.Add((UIDynamic) tmpColor);
                
                SaberOpt_ColorTendrilsB = new JSONStorableColor("SaberOpt_ColorTendrilsB", hsvcDefaultBlade);
                SaberOpt_ColorTendrilsB.setJSONCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpColor = CreateColorPicker(SaberOpt_ColorTendrilsB, false);
                tmpColor.label = "Tendrils Color";
                secSaberColorOptsUIs.Add((UIDynamic) tmpColor);
                
                tmpSpacer = CreateSpacer(false);
                Utils.SetupSpacer(tmpSpacer, 20f);
                secSaberColorOptsUIs.Add((UIDynamic) tmpSpacer);
                
                SaberOpt_ColorEmitterB = new JSONStorableColor("SaberOpt_ColorEmitterB", hsvcDefault);
                SaberOpt_ColorEmitterB.setJSONCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpColor = CreateColorPicker(SaberOpt_ColorEmitterB, false);
                tmpColor.label = "Emitter Color";
                secSaberColorOptsUIs.Add((UIDynamic) tmpColor);
                
                SaberOpt_ColorSwitchB = new JSONStorableColor("SaberOpt_ColorSwitchB", hsvcDefault);
                SaberOpt_ColorSwitchB.setJSONCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpColor = CreateColorPicker(SaberOpt_ColorSwitchB, false);
                tmpColor.label = "Switch Color";
                secSaberColorOptsUIs.Add((UIDynamic) tmpColor);
                
                SaberOpt_ColorHandleB = new JSONStorableColor("SaberOpt_ColorHandleB", hsvcDefault);
                SaberOpt_ColorHandleB.setJSONCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpColor = CreateColorPicker(SaberOpt_ColorHandleB, false);
                tmpColor.label = "Handle Color";
                secSaberColorOptsUIs.Add((UIDynamic) tmpColor);

                SaberOpt_ColorPowerCellB = new JSONStorableColor("SaberOpt_ColorPowerCellB", hsvcDefault);
                SaberOpt_ColorPowerCellB.setJSONCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpColor = CreateColorPicker(SaberOpt_ColorPowerCellB, false);
                tmpColor.label = "Power cell Color";
                secSaberColorOptsUIs.Add((UIDynamic) tmpColor);
                
                /* ********************************* */
                /* **** RIGHT UI                **** */
                /* ********************************* */
                Utils.CreateStaticDescriptionText(this,"VAMLightsabers","<color=#000><size=35><b>VAMLightsabers</b></size></color>",true,40,TextAnchor.MiddleRight);

                RightUISelector = new JSONStorableStringChooser("Right UI Selector", uisChoices, uisChoices.First(), "Show") {isStorable=false,isRestorable=false};
                RightUISelector.setCallbackFunction += (val) => { OnChangeRightUISelection(); };
                UIDynamicPopup RUISelsp = CreateScrollablePopup(RightUISelector, true);
                RUISelsp.labelWidth = 125f;

                tmpSpacer = CreateSpacer(true);
                Utils.SetupSpacer(tmpSpacer, 20f);

                // **** Main saber options ****
                tmpTextfield = Utils.CreateStaticDescriptionText(this,"PartsMain","<color=#000><size=28><b>Main lightsaber parts</b></size></color>\n<size=28>Select your parts for the main lightsaber (or one handed lightsaber).</size>",true,95,TextAnchor.MiddleRight);
                mainSaberOptsUIs.Add((UIDynamic) tmpTextfield);
                
                SaberOpt_BladeAOffset = new JSONStorableFloat("SaberOpt_BladeAOffset", -0.025f, -0.1f, 0.1f);
                SaberOpt_BladeAOffset.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpSlider = CreateSlider(SaberOpt_BladeAOffset, true);
                tmpSlider.valueFormat = "F3";
                tmpSlider.label = "Blade Offset";
                mainSaberOptsUIs.Add((UIDynamic) tmpSlider);
                
                SaberOpt_BladeA = new JSONStorableStringChooser("SaberOpt_BladeA", _bladesChoices, "", "Blade type");
                SaberOpt_BladeA.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpPopup = CreateScrollablePopup(SaberOpt_BladeA, true);
                tmpPopup.labelWidth = 150f;
                mainSaberOptsUIs.Add((UIDynamic) tmpPopup);
                
                SaberOpt_BladeAEnableLightning = new JSONStorableBool("SaberOpt_BladeAEnableLightning", false);
                SaberOpt_BladeAEnableLightning.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpToggle = CreateToggle(SaberOpt_BladeAEnableLightning, true);
                tmpToggle.label = "Enable lightning";
                mainSaberOptsUIs.Add((UIDynamic) tmpToggle);
                
                SaberOpt_BladeAEnableTendrils = new JSONStorableBool("SaberOpt_BladeAEnableTendrils", false);
                SaberOpt_BladeAEnableTendrils.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpToggle = CreateToggle(SaberOpt_BladeAEnableTendrils, true);
                tmpToggle.label = "Enable tendrils";
                mainSaberOptsUIs.Add((UIDynamic) tmpToggle);
                
                SaberOpt_EmitterA = new JSONStorableStringChooser("SaberOpt_EmitterA", _hiltsChoices, DEFAULT_SKIN, "Emitter");
                SaberOpt_EmitterA.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpPopup = CreateScrollablePopup(SaberOpt_EmitterA, true);
                tmpPopup.labelWidth = 150f;
                mainSaberOptsUIs.Add((UIDynamic) tmpPopup);
                
                SaberOpt_SwitchA = new JSONStorableStringChooser("SaberOpt_SwitchA", _hiltsChoices, DEFAULT_SKIN, "Switch");
                SaberOpt_SwitchA.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpPopup = CreateScrollablePopup(SaberOpt_SwitchA, true);
                tmpPopup.labelWidth = 150f;
                mainSaberOptsUIs.Add((UIDynamic) tmpPopup);
                
                SaberOpt_HandleA = new JSONStorableStringChooser("SaberOpt_HandleA", _hiltsChoices, DEFAULT_SKIN, "Handle");
                SaberOpt_HandleA.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpPopup = CreateScrollablePopup(SaberOpt_HandleA, true);
                tmpPopup.labelWidth = 150f;
                mainSaberOptsUIs.Add((UIDynamic) tmpPopup);

                SaberOpt_PowerCellA = new JSONStorableStringChooser("SaberOpt_PowerCellA", _hiltsChoices, DEFAULT_SKIN, "Power Cell");
                SaberOpt_PowerCellA.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpPopup = CreateScrollablePopup(SaberOpt_PowerCellA, true);
                tmpPopup.labelWidth = 150f;
                mainSaberOptsUIs.Add((UIDynamic) tmpPopup);

                tmpSpacer = CreateSpacer(true);
                Utils.SetupSpacer(tmpSpacer, 20f);
                mainSaberOptsUIs.Add((UIDynamic) tmpSpacer);
                
                tmpTextfield = Utils.CreateStaticDescriptionText(this,"PartsMain","<color=#000><size=28><b>Main lightsaber offsets</b></size></color>\n<size=28>Change the vertical offset for the parts.</size>",true,95,TextAnchor.MiddleRight);
                mainSaberOptsUIs.Add((UIDynamic) tmpTextfield);
                
                SaberOpt_SwitchAOffset = new JSONStorableFloat("SaberOpt_SwitchAOffset", 0f, -0.1f, 0.1f);
                SaberOpt_SwitchAOffset.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpSlider = CreateSlider(SaberOpt_SwitchAOffset, true);
                tmpSlider.valueFormat = "F4";
                tmpSlider.label = "Switch Offset";
                mainSaberOptsUIs.Add((UIDynamic) tmpSlider);
                
                SaberOpt_HandleAOffset = new JSONStorableFloat("SaberOpt_HandleAOffset", 0f, -0.1f, 0.1f);
                SaberOpt_HandleAOffset.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpSlider = CreateSlider(SaberOpt_HandleAOffset, true);
                tmpSlider.valueFormat = "F4";
                tmpSlider.label = "Handle Offset";
                mainSaberOptsUIs.Add((UIDynamic) tmpSlider);
                
                SaberOpt_PowerCellAOffset = new JSONStorableFloat("SaberOpt_PowerCellAOffset", 0f, -0.1f, 0.1f);
                SaberOpt_PowerCellAOffset.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpSlider = CreateSlider(SaberOpt_PowerCellAOffset, true);
                tmpSlider.valueFormat = "F4";
                tmpSlider.label = "Power Cell Offset";
                mainSaberOptsUIs.Add((UIDynamic) tmpSlider);

                // **** Secondary saber options ****
                tmpTextfield = Utils.CreateStaticDescriptionText(this,"PartsSec","<color=#000><size=28><b>Secondary lightsaber parts</b></size></color>\n<size=28>Enable and select your parts for the secondary lightsaber.</size>",true,95,TextAnchor.MiddleRight);
                secSaberOptsUIs.Add((UIDynamic) tmpTextfield);
                
                SaberOpt_EnableDualBladed = new JSONStorableBool("SaberOpt_EnableDualBladed", false);
                SaberOpt_EnableDualBladed.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpToggle = CreateToggle(SaberOpt_EnableDualBladed, true);
                tmpToggle.label = "Enable dual bladed saber";
                secSaberOptsUIs.Add((UIDynamic) tmpToggle);
                
                SaberOpt_DualBladedOnlyOneSwitch = new JSONStorableBool("SaberOpt_DualBladedOnlyOneSwitch", false);
                SaberOpt_DualBladedOnlyOneSwitch.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpToggle = CreateToggle(SaberOpt_DualBladedOnlyOneSwitch, true);
                tmpToggle.label = "Only one switch";
                secSaberOptsUIs.Add((UIDynamic) tmpToggle);
                
                SaberOpt_DualBladeHandleOffset = new JSONStorableFloat("SaberOpt_DualBladeHandleOffset", -0.004f, -0.1f, 0.1f);
                SaberOpt_DualBladeHandleOffset.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpSlider = CreateSlider(SaberOpt_DualBladeHandleOffset, true);
                tmpSlider.valueFormat = "F3";
                tmpSlider.label = "Secondary handle offset";
                secSaberOptsUIs.Add((UIDynamic) tmpSlider);
                
                SaberOpt_BladeB = new JSONStorableStringChooser("SaberOpt_BladeB", _bladesChoices, "", "Blade type");
                SaberOpt_BladeB.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpPopup = CreateScrollablePopup(SaberOpt_BladeB, true);
                tmpPopup.labelWidth = 150f;
                secSaberOptsUIs.Add((UIDynamic) tmpPopup);
                
                SaberOpt_BladeBEnableLightning = new JSONStorableBool("SaberOpt_BladeBEnableLightning", false);
                SaberOpt_BladeBEnableLightning.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpToggle = CreateToggle(SaberOpt_BladeBEnableLightning, true);
                tmpToggle.label = "Enable lightning";
                secSaberOptsUIs.Add((UIDynamic) tmpToggle);
                
                SaberOpt_BladeBEnableTendrils = new JSONStorableBool("SaberOpt_BladeBEnableTendrils", false);
                SaberOpt_BladeBEnableTendrils.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpToggle = CreateToggle(SaberOpt_BladeBEnableTendrils, true);
                tmpToggle.label = "Enable tendrils";
                secSaberOptsUIs.Add((UIDynamic) tmpToggle);
                
                SaberOpt_DualBladeHandleRotation = new JSONStorableFloat("SaberOpt_DualBladeHandleRotation", 0f, 0f, 360f);
                SaberOpt_DualBladeHandleRotation.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpSlider = CreateSlider(SaberOpt_DualBladeHandleRotation, true);
                tmpSlider.valueFormat = "F0";
                tmpSlider.label = "Secondary handle rotation";
                secSaberOptsUIs.Add((UIDynamic) tmpSlider);
                
                SaberOpt_DualBladedCellCount = new JSONStorableStringChooser("SaberOpt_DualBladedCellCount", _cellCountChoices, "2", "Power Cell Count");
                SaberOpt_DualBladedCellCount.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpPopup = CreateScrollablePopup(SaberOpt_DualBladedCellCount, true);
                tmpPopup.labelWidth = 250f;
                secSaberOptsUIs.Add((UIDynamic) tmpPopup);

                SaberOpt_BladeBOffset = new JSONStorableFloat("SaberOpt_BladeBOffset", -0.008f, -0.1f, 0.1f);
                SaberOpt_BladeBOffset.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpSlider = CreateSlider(SaberOpt_BladeBOffset, true);
                tmpSlider.valueFormat = "F3";
                tmpSlider.label = "Blade Offset";
                secSaberOptsUIs.Add((UIDynamic) tmpSlider);
                
                SaberOpt_EmitterB = new JSONStorableStringChooser("SaberOpt_EmitterB", _hiltsChoices, "", "Emitter");
                SaberOpt_EmitterB.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpPopup = CreateScrollablePopup(SaberOpt_EmitterB, true);
                tmpPopup.labelWidth = 150f;
                secSaberOptsUIs.Add((UIDynamic) tmpPopup);
                
                SaberOpt_SwitchB = new JSONStorableStringChooser("SaberOpt_SwitchB", _hiltsChoices, "", "Switch");
                SaberOpt_SwitchB.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpPopup = CreateScrollablePopup(SaberOpt_SwitchB, true);
                tmpPopup.labelWidth = 150f;
                secSaberOptsUIs.Add((UIDynamic) tmpPopup);
                
                SaberOpt_HandleB = new JSONStorableStringChooser("SaberOpt_HandleB", _hiltsChoices, "", "Handle");
                SaberOpt_HandleB.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpPopup = CreateScrollablePopup(SaberOpt_HandleB, true);
                tmpPopup.labelWidth = 150f;
                secSaberOptsUIs.Add((UIDynamic) tmpPopup);

                SaberOpt_PowerCellB = new JSONStorableStringChooser("SaberOpt_PowerCellB", _hiltsChoices, "", "Power Cell");
                SaberOpt_PowerCellB.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpPopup = CreateScrollablePopup(SaberOpt_PowerCellB, true);
                tmpPopup.labelWidth = 150f;
                secSaberOptsUIs.Add((UIDynamic) tmpPopup);
                
                tmpSpacer = CreateSpacer(true);
                Utils.SetupSpacer(tmpSpacer, 20f);
                secSaberOptsUIs.Add((UIDynamic) tmpSpacer);
                
                tmpTextfield = Utils.CreateStaticDescriptionText(this,"PartsMain","<color=#000><size=28><b>Secondary lightsaber offsets</b></size></color>\n<size=28>Change the vertical offset for the parts.</size>",true,95,TextAnchor.MiddleRight);
                secSaberOptsUIs.Add((UIDynamic) tmpTextfield);
                
                SaberOpt_SwitchBOffset = new JSONStorableFloat("SaberOpt_SwitchBOffset", 0f, -0.1f, 0.1f);
                SaberOpt_SwitchBOffset.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpSlider = CreateSlider(SaberOpt_SwitchBOffset, true);
                tmpSlider.valueFormat = "F4";
                tmpSlider.label = "Switch Offset";
                secSaberOptsUIs.Add((UIDynamic) tmpSlider);
                
                SaberOpt_HandleBOffset = new JSONStorableFloat("SaberOpt_HandleBOffset", 0f, -0.1f, 0.1f);
                SaberOpt_HandleBOffset.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpSlider = CreateSlider(SaberOpt_HandleBOffset, true);
                tmpSlider.valueFormat = "F4";
                tmpSlider.label = "Handle Offset";
                secSaberOptsUIs.Add((UIDynamic) tmpSlider);
                
                SaberOpt_PowerCellBOffset = new JSONStorableFloat("SaberOpt_PowerCellBOffset", 0f, -0.1f, 0.1f);
                SaberOpt_PowerCellBOffset.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpSlider = CreateSlider(SaberOpt_PowerCellBOffset, true);
                tmpSlider.valueFormat = "F4";
                tmpSlider.label = "Power Cell Offset";
                secSaberOptsUIs.Add((UIDynamic) tmpSlider);
                
                // **** Global saber options ****
                tmpTextfield = Utils.CreateStaticDescriptionText(this,"GlobalSaber","<color=#000><size=28><b>Global saber options</b></size></color>\n<size=28>Tweak global options for the whole saber.</size>",true,95,TextAnchor.MiddleRight);
                globalSaberOptsUIs.Add((UIDynamic) tmpTextfield);

                SaberOpt_StartsOn = new JSONStorableBool("SaberOpt_StartsOn", false);
                SaberOpt_StartsOn.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpToggle = CreateToggle(SaberOpt_StartsOn, true);
                tmpToggle.label = "Saber starts on";
                globalSaberOptsUIs.Add((UIDynamic) tmpToggle);
                
                SaberOpt_SwitchSoundsVolume = new JSONStorableFloat("SaberOpt_SwitchSoundsVolume", 0.8f, 0f, 1f);
                SaberOpt_SwitchSoundsVolume.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpSlider = CreateSlider(SaberOpt_SwitchSoundsVolume, true);
                tmpSlider.valueFormat = "F2";
                tmpSlider.label = "Switch sounds volume";
                globalSaberOptsUIs.Add((UIDynamic) tmpSlider);
                
                SaberOpt_LoopSoundsVolume = new JSONStorableFloat("SaberOpt_LoopSoundsVolume", 0.8f, 0f, 1f);
                SaberOpt_LoopSoundsVolume.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpSlider = CreateSlider(SaberOpt_LoopSoundsVolume, true);
                tmpSlider.valueFormat = "F2";
                tmpSlider.label = "Loop sounds volume";
                globalSaberOptsUIs.Add((UIDynamic) tmpSlider);
                
                SaberOpt_OnSound = new JSONStorableStringChooser("SaberOpt_OnSound", _onSndChoices, "on_01", "On sound");
                SaberOpt_OnSound.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpPopup = CreateScrollablePopup(SaberOpt_OnSound, true);
                tmpPopup.labelWidth = 250f;
                globalSaberOptsUIs.Add((UIDynamic) tmpPopup);
                
                SaberOpt_OnSoundRandom = new JSONStorableBool("SaberOpt_OnSoundRandom", false);
                SaberOpt_OnSoundRandom.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpToggle = CreateToggle(SaberOpt_OnSoundRandom, true);
                tmpToggle.label = "Choose On sound randomly";
                globalSaberOptsUIs.Add((UIDynamic) tmpToggle);
                
                SaberOpt_OffSound = new JSONStorableStringChooser("SaberOpt_OffSound", _offSndChoices, "off_01", "Off sound");
                SaberOpt_OffSound.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpPopup = CreateScrollablePopup(SaberOpt_OffSound, true);
                tmpPopup.labelWidth = 250f;
                globalSaberOptsUIs.Add((UIDynamic) tmpPopup);
                
                SaberOpt_OffSoundRandom = new JSONStorableBool("SaberOpt_OffSoundRandom", false);
                SaberOpt_OffSoundRandom.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpToggle = CreateToggle(SaberOpt_OffSoundRandom, true);
                tmpToggle.label = "Choose Off sound randomly";
                globalSaberOptsUIs.Add((UIDynamic) tmpToggle);
                
                SaberOpt_LoopSound = new JSONStorableStringChooser("SaberOpt_LoopSound", _loopSndChoices, "loop_01", "Loop sound");
                SaberOpt_LoopSound.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpPopup = CreateScrollablePopup(SaberOpt_LoopSound, true);
                tmpPopup.labelWidth = 250f;
                globalSaberOptsUIs.Add((UIDynamic) tmpPopup);
                
                SaberOpt_BladesFresnelScale = new JSONStorableFloat("SaberOpt_BladesFresnelScale", 2f, 0f, 10f);
                SaberOpt_BladesFresnelScale.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpSlider = CreateSlider(SaberOpt_BladesFresnelScale, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Blades color fresnel scale";
                globalSaberOptsUIs.Add((UIDynamic) tmpSlider);
                
                SaberOpt_BladesFresnelPower = new JSONStorableFloat("SaberOpt_BladesFresnelPower", 6f, 0f, 50f);
                SaberOpt_BladesFresnelPower.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpSlider = CreateSlider(SaberOpt_BladesFresnelPower, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Blades color fresnel power";
                globalSaberOptsUIs.Add((UIDynamic) tmpSlider);
                
                SaberOpt_PartsRoughness = new JSONStorableFloat("SaberOpt_PartsRoughness", 1f, 0f, 1f);
                SaberOpt_PartsRoughness.setCallbackFunction += (val) => { OnChangeLightsaberOptions(); };
                tmpSlider = CreateSlider(SaberOpt_PartsRoughness, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Parts roughness";
                globalSaberOptsUIs.Add((UIDynamic) tmpSlider);
                
                /* ***************************************** */
                /* **** ACTIONS                         **** */
                /* ***************************************** */
                JSONStorableAction fakeFuncUseBelow = new JSONStorableAction("- - - - Use these functions below ↓ - - - - -", () => {});
                RegisterAction(fakeFuncUseBelow);
                
                A_LS_SwitchOn = new JSONStorableAction("Lightsaber Switch On", () => { OnLightsaberSwitchOn(); });
                RegisterAction(A_LS_SwitchOn);
                
                A_LS_SwitchOff = new JSONStorableAction("Lightsaber Switch Off", () => { OnLightsaberSwitchOff(); });
                RegisterAction(A_LS_SwitchOff);
                
                /* ***************************************** */
                /* **** REGISTERING STORABLES   **** */
                /* ***************************************** */
                JSONStorableAction fakeFuncStorables = new JSONStorableAction("- - - - - - - - - - - - - - - - - - -", () => {});
                RegisterAction(fakeFuncStorables);
                
                RegisterColor(SaberOpt_ColorBladeA);
                RegisterColor(SaberOpt_ColorBladeGlowA);
                RegisterColor(SaberOpt_ColorBladeGlowA);
                RegisterColor(SaberOpt_ColorLightningA);
                RegisterColor(SaberOpt_ColorTendrilsA);
                RegisterColor(SaberOpt_ColorEmitterA);
                RegisterColor(SaberOpt_ColorSwitchA);
                RegisterColor(SaberOpt_ColorHandleA);
                RegisterColor(SaberOpt_ColorPowerCellA);
                
                RegisterColor(SaberOpt_ColorBladeB);
                RegisterColor(SaberOpt_ColorBladeGlowB);
                RegisterColor(SaberOpt_ColorBladeGlowB);
                RegisterColor(SaberOpt_ColorLightningB);
                RegisterColor(SaberOpt_ColorTendrilsB);
                RegisterColor(SaberOpt_ColorEmitterB);
                RegisterColor(SaberOpt_ColorSwitchB);
                RegisterColor(SaberOpt_ColorHandleB);
                RegisterColor(SaberOpt_ColorPowerCellB);

                RegisterFloat(SaberOpt_BladeAOffset);
                RegisterStringChooser(SaberOpt_BladeA);
                RegisterBool(SaberOpt_BladeAEnableLightning);
                RegisterBool(SaberOpt_BladeAEnableTendrils);
                RegisterStringChooser(SaberOpt_EmitterA);
                RegisterStringChooser(SaberOpt_SwitchA);
                RegisterStringChooser(SaberOpt_HandleA);
                RegisterStringChooser(SaberOpt_PowerCellA);
                RegisterFloat(SaberOpt_SwitchAOffset);
                RegisterFloat(SaberOpt_HandleAOffset);
                RegisterFloat(SaberOpt_PowerCellAOffset);
                
                RegisterFloat(SaberOpt_BladeBOffset);
                RegisterStringChooser(SaberOpt_BladeB);
                RegisterBool(SaberOpt_BladeBEnableLightning);
                RegisterBool(SaberOpt_BladeBEnableTendrils);
                RegisterStringChooser(SaberOpt_EmitterB);
                RegisterStringChooser(SaberOpt_SwitchB);
                RegisterStringChooser(SaberOpt_HandleB);
                RegisterStringChooser(SaberOpt_PowerCellB);
                RegisterFloat(SaberOpt_SwitchBOffset);
                RegisterFloat(SaberOpt_HandleBOffset);
                RegisterFloat(SaberOpt_PowerCellBOffset);

                RegisterBool(SaberOpt_EnableDualBladed);
                RegisterBool(SaberOpt_DualBladedOnlyOneSwitch);
                RegisterStringChooser(SaberOpt_DualBladedCellCount);
                RegisterFloat(SaberOpt_DualBladeHandleOffset);
                RegisterFloat(SaberOpt_DualBladeHandleRotation);
                
                RegisterBool(SaberOpt_StartsOn);
                RegisterFloat(SaberOpt_SwitchSoundsVolume);
                RegisterFloat(SaberOpt_LoopSoundsVolume);
                
                RegisterStringChooser(SaberOpt_OnSound);
                RegisterBool(SaberOpt_OnSoundRandom);
                RegisterStringChooser(SaberOpt_OffSound);
                RegisterBool(SaberOpt_OffSoundRandom);
                RegisterStringChooser(SaberOpt_LoopSound);
                
                RegisterFloat(SaberOpt_BladesFresnelScale);
                RegisterFloat(SaberOpt_BladesFresnelPower);
                RegisterFloat(SaberOpt_PartsRoughness);
                
                VAMLSValue_Ready = new JSONStorableBool("VAMLSValue_Ready", false) {isStorable=false,isRestorable=false};
                RegisterBool(VAMLSValue_Ready);

                // Init UI
                OnChangeLeftUISelection();
                OnChangeRightUISelection();
                
                // Loading the assetbundle
				Request request = new AssetLoader.AssetBundleFromFileRequest {path = PLUGIN_PATH + SABER_BUNDLE_PATH, callback = OnBundleLoaded};
				AssetLoader.QueueLoadAssetBundleFromFile(request);
            }
            catch(Exception e)
            {
                SuperController.LogError(EXCEPTION_PREFIX + e);
            }
		}
		
		void Update()
		{
			if (_isSceneLoading || _isSceneFrozen) return;

			if (!_isSceneLoading && _isBundleLoaded && _vamlsInitialized == false)
			{

				_vamlsInitialized = true;
				VAMLSValue_Ready.val = true;
				
				OnChangeLightsaberOptions();
				
				// If the saber starts on
				if (SaberOpt_StartsOn.val == true)
				{
					_ANIMSaber.Play("LightsaberOn");
				}
				else
				{
					_ANIMSaber.Play("LightsaberOff");
				}
				
				// Setting the default values on the empty
				containingAtom.SetBoolParamValue("collisionEnabled", true); // Collision for the hilt
				containingAtom.GetStorableByID("scale").SetFloatParamValue("scale", 0.01f); // Scale for the empty
				
				// Setting a default name for the saber if it's the default "Empty" name
				// the setUID will properly add a #X if it's already in use
				if (containingAtom.name == "Empty")
				{
					containingAtom.SetUID("VAMLightsaber");
				}
			}
			
			// No update if not ready
			if (!_vamlsInitialized) return;

			// Internal timer
			_vamexmTimer += Time.deltaTime;
		}

		// **************************
		// Functions
		// **************************
		protected void CleanHiltParent()
		{
			if (!_vamlsInitialized) return;
			foreach (Transform child in _TRHiltParent)
			{
				Destroy(child.gameObject);
			}
		}
		
		protected void InstantiatePart(GameObject go, JSONStorableColor color, bool dualBladeComponent = false, bool dualBladeCompFirst = false, float compOffset = 0f)
		{
			if (!_vamlsInitialized) return;
			
			GameObject partGO;
			MeshFilter partMeshFilter;
			Renderer partRenderer;
			float partPosition = 0f;
			
			partGO = Instantiate(go);
			partGO.transform.SetParent(_TRHiltParent, false);

			partMeshFilter = partGO.GetComponent<MeshFilter>();
			partRenderer = partGO.GetComponent<Renderer>();
			Color newPartColor = new Color(color.colorPicker.currentColor.r, color.colorPicker.currentColor.g, color.colorPicker.currentColor.b, 1f);
			partRenderer.material.SetColor("_ColorCorrection", newPartColor);
			
			if (dualBladeComponent == false)
			{
				_partBoundsHeight -= partMeshFilter.mesh.bounds.size.y + compOffset;
				partGO.transform.localPosition = new Vector3(0, _partBoundsHeight, 0);
				partGO.transform.localRotation = Quaternion.Euler(0f, 0f, 0f);
			}
			else
			{
				if (dualBladeCompFirst)
				{
					_partBoundsHeight -= SaberOpt_DualBladeHandleOffset.val;
				}

				partGO.transform.localPosition = new Vector3(0, _partBoundsHeight, 0);
				partGO.transform.localRotation = Quaternion.Euler(0f, SaberOpt_DualBladeHandleRotation.val, 180f);
				
				_partBoundsHeight -= partMeshFilter.mesh.bounds.size.y + compOffset;
			}
		}

		protected void UpdatePartsRoughness()
		{
			foreach (Transform trPart in _TRHiltParent)
			{
				Renderer partRN = trPart.GetComponent<Renderer>();
				partRN.material.SetFloat("_RoughnessIntensity", SaberOpt_PartsRoughness.val);
			}
		}

		protected void SelectBladeType(Transform typeParent, string typeName, bool secondaryBlade = false)
		{
			if (!_vamlsInitialized) return;
			
			foreach (Transform trType in typeParent.GetChild(0))
			{
				bool stateBlade = trType.name == typeName ? true : false;
				trType.gameObject.SetActive(stateBlade);
				if (stateBlade == true)
				{
					bool lightningState;
					bool tendrilsState;
					JSONStorableColor bladeCol;
					JSONStorableColor bladeGlowCol;
					JSONStorableColor lightningCol;
					JSONStorableColor tendrilsCol;
					if (secondaryBlade)
					{
						bladeCol = SaberOpt_ColorBladeB;
						bladeGlowCol = SaberOpt_ColorBladeGlowB;
						lightningState = SaberOpt_BladeBEnableLightning.val;
						lightningCol = SaberOpt_ColorLightningB;
						tendrilsState = SaberOpt_BladeBEnableTendrils.val;
						tendrilsCol = SaberOpt_ColorTendrilsB;
					}
					else
					{
						bladeCol = SaberOpt_ColorBladeA;
						bladeGlowCol = SaberOpt_ColorBladeGlowA;
						lightningState = SaberOpt_BladeAEnableLightning.val;
						lightningCol = SaberOpt_ColorLightningA;
						tendrilsState = SaberOpt_BladeAEnableTendrils.val;
						tendrilsCol = SaberOpt_ColorTendrilsA;
					}

					ChangeBladeColor(trType,bladeCol);
					
					Transform trGlowVFX = Utils.GetChildNamed(trType, "blade_glow");
					ChangeVFXColor(trGlowVFX, bladeGlowCol);
					
					Transform trLightningVFX = Utils.GetChildNamed(trType, "blade_vfx_lightning");
					trLightningVFX.gameObject.SetActive(lightningState);
					if (lightningState) ChangeVFXColor(trLightningVFX, lightningCol);
					
					Transform trTendrilsVFX = Utils.GetChildNamed(trType, "blade_vfx_tendrils");
					trTendrilsVFX.gameObject.SetActive(tendrilsState);
					if (tendrilsState) ChangeVFXColor(trTendrilsVFX, tendrilsCol);
				}
			}
		}

		protected void ChangeBladeColor(Transform bladeTr, JSONStorableColor color)
		{
			Color newColor = new Color(color.colorPicker.currentColor.r, color.colorPicker.currentColor.g, color.colorPicker.currentColor.b, 1f);
			Renderer bladeRN = bladeTr.GetComponent<Renderer>();
			bladeRN.material.SetColor("_BladeColor", newColor);
			bladeRN.material.SetFloat("_FresnelScale", SaberOpt_BladesFresnelScale.val);
			bladeRN.material.SetFloat("_FresnelPower", SaberOpt_BladesFresnelPower.val);
		}
		
		protected void ChangeVFXColor(Transform vfxTR, JSONStorableColor color)
		{
			Color newColor = new Color(color.colorPicker.currentColor.r, color.colorPicker.currentColor.g, color.colorPicker.currentColor.b, 1f);
			ParticleSystem prtComponent = vfxTR.GetComponent<ParticleSystem>();
			var prtMain = prtComponent.main;
			prtMain.startColor = newColor;
		}

		protected void UpdateBladeTypesChoices(JSONStorableStringChooser typeChooser, List<string> choices)
		{
			typeChooser.choices = choices;
			if (typeChooser.val == "") typeChooser.valNoCallback = choices.First();
		}
			
		protected void UpdatePartsChoices(JSONStorableStringChooser partChooser, List<string> choices )
		{
			partChooser.choices = choices;
			if (partChooser.val == "") partChooser.valNoCallback = choices.First();
		}

		protected void LightsaberSwitchOn()
		{
			UpdateSoundBeforeSwitchingOn();
			_ANIMSaber.Play("LightsaberSwitchOn");
		}
		
		protected void LightsaberSwitchOff()
		{
			_ANIMSaber.Play("LightsaberSwitchOff");
		}

		protected void UpdateSoundBeforeSwitchingOn()
		{
			// Switch On Sound
			if ( SaberOpt_OnSoundRandom.val == true )
			{
				int sndChoice = UnityEngine.Random.Range(0, _onSndSounds.Count);
				_ASOnSnd.clip = _onSndSounds[sndChoice];
			}
			else
			{
				_ASOnSnd.clip = _onSndSounds[_onSndChoices.IndexOf(SaberOpt_OnSound.val)];
			}
			
			// Switch Off Sound
			if ( SaberOpt_OffSoundRandom.val == true )
			{
				int sndChoice = UnityEngine.Random.Range(0, _offSndSounds.Count);
				_ASOffSnd.clip = _offSndSounds[sndChoice];
			}
			else
			{
				_ASOffSnd.clip = _offSndSounds[_offSndChoices.IndexOf(SaberOpt_OffSound.val)];
			}
			
			// Loop sound
			_ASLoopSnd.clip = _loopSndSounds[_loopSndChoices.IndexOf(SaberOpt_LoopSound.val)];
		}
		
		// **************************************
		// Callbacks & similar ( internal )
		// **************************************
		protected void OnChangeLeftUISelection()
		{
			bool mainSaberVisible = LeftUISelector.val == "Main Saber Colors" ? true : false;
			bool secSaberVisible = LeftUISelector.val == "Secondary Saber Colors" ? true : false;

			ToggleUIDynamicGrp(mainSaberColorOptsUIs, mainSaberVisible);
			ToggleUIDynamicGrp(secSaberColorOptsUIs, secSaberVisible);
		}
		
		protected void OnChangeRightUISelection()
		{
			bool mainSaberVisible = RightUISelector.val == "Main Saber Options" ? true : false;
			bool secSaberVisible = RightUISelector.val == "Secondary Saber Options" ? true : false;
			bool globalSaberVisible = RightUISelector.val == "Global Saber Options" ? true : false;

			ToggleUIDynamicGrp(mainSaberOptsUIs, mainSaberVisible);
			ToggleUIDynamicGrp(secSaberOptsUIs, secSaberVisible);
			ToggleUIDynamicGrp(globalSaberOptsUIs, globalSaberVisible);
		}
		
		protected void OnChangeLightsaberOptions()
		{
			if (!_vamlsInitialized) return;
			
			CleanHiltParent();

			GameObject selectedPart;
			_partBoundsHeight = 0f;
			int pcellCount = int.Parse(SaberOpt_DualBladedCellCount.val);

			selectedPart = _hiltsGameobjects["emitter"][_hiltsChoices.IndexOf(SaberOpt_EmitterA.val)];
			InstantiatePart( selectedPart, SaberOpt_ColorEmitterA );

			selectedPart = _hiltsGameobjects["switch"][ _hiltsChoices.IndexOf(SaberOpt_SwitchA.val) ];
			InstantiatePart( selectedPart, SaberOpt_ColorSwitchA, false, false, SaberOpt_SwitchAOffset.val );
			
			selectedPart = _hiltsGameobjects["handle"][_hiltsChoices.IndexOf(SaberOpt_HandleA.val)];
			InstantiatePart( selectedPart, SaberOpt_ColorHandleA, false, false, SaberOpt_HandleAOffset.val );
			
			if (SaberOpt_EnableDualBladed.val == false || ( SaberOpt_EnableDualBladed.val == true && pcellCount >= 1 ) )
			{
				selectedPart = _hiltsGameobjects["powercell"][_hiltsChoices.IndexOf(SaberOpt_PowerCellA.val)];
				InstantiatePart(selectedPart, SaberOpt_ColorPowerCellA, false, false, SaberOpt_PowerCellAOffset.val);
			}
			
			// First blade offset
			_TRBladeParent.localPosition = new Vector3(0, SaberOpt_BladeAOffset.val, 0);
			
			// Blade type
			SelectBladeType(_TRBladeParent, SaberOpt_BladeA.val);

			// Dual bladed
			if ( SaberOpt_EnableDualBladed.val == true )
			{
				bool firstPartPcell = pcellCount == 2 ? true : false;
				bool firstPartHandle = pcellCount < 2 && SaberOpt_DualBladedOnlyOneSwitch.val == true ? true : false;

				if (pcellCount == 2)
				{
					selectedPart = _hiltsGameobjects["powercell"][ _hiltsChoices.IndexOf(SaberOpt_PowerCellB.val) ];
					InstantiatePart( selectedPart, SaberOpt_ColorPowerCellB, true, firstPartPcell, SaberOpt_PowerCellBOffset.val );					
				}

				selectedPart = _hiltsGameobjects["handle"][_hiltsChoices.IndexOf(SaberOpt_HandleB.val)];
				InstantiatePart( selectedPart, SaberOpt_ColorHandleB, true, firstPartHandle, SaberOpt_HandleBOffset.val );

				if (SaberOpt_DualBladedOnlyOneSwitch.val == false)
				{
					selectedPart = _hiltsGameobjects["switch"][_hiltsChoices.IndexOf(SaberOpt_SwitchB.val)];
					InstantiatePart(selectedPart, SaberOpt_ColorSwitchB, true, false, SaberOpt_SwitchBOffset.val);
				}
				
				selectedPart = _hiltsGameobjects["emitter"][_hiltsChoices.IndexOf(SaberOpt_EmitterB.val)];
				InstantiatePart( selectedPart, SaberOpt_ColorEmitterB, true );
				
				// Enabling blade and offset
				_TRBlade2Parent.gameObject.SetActive(true);
				_TRBlade2Parent.localPosition = new Vector3(0, _partBoundsHeight - SaberOpt_BladeBOffset.val, 0);
				
				// Blade type
				SelectBladeType(_TRBlade2Parent, SaberOpt_BladeB.val, true);
			}
			else
			{
				// Disabling dual blade
				_TRBlade2Parent.gameObject.SetActive(false);
			}
			
			// Updating materials roughness
			UpdatePartsRoughness();
			
			// Centering the saber to the gizmo
			_TRSaberParent.localPosition = new Vector3(0, -(_partBoundsHeight/2), 0);
			
			// Volumes
			_ASLoopSnd.volume = SaberOpt_LoopSoundsVolume.val;
			_ASOnSnd.volume = SaberOpt_SwitchSoundsVolume.val;
			_ASOffSnd.volume = SaberOpt_SwitchSoundsVolume.val;
		}
		
		protected void OnBundleLoaded(Request aRequest) {
			try
			{
				// Reading my bundle
				bundleRequest = aRequest;
				GameObject tmp = bundleRequest.assetBundle.LoadAsset<GameObject>("assets/vamlightsabers/prefabs/vamlightsaber.prefab");
				VAMLSGO = Instantiate(tmp);
				
				// Parenting to my current Empty
				VAMLSGO.transform.SetParent(containingAtom.reParentObject.GetChild(0).transform, false);
				VAMLSGO.transform.localPosition = new Vector3(0, 0, 0);
				VAMLSGO.transform.localRotation = Quaternion.Euler(0f,0f,0f);
				
				_TRSaberMain = VAMLSGO.transform;
				_TRSaberParent = Utils.GetChildNamed(_TRSaberMain, "saber_parent");
				_TRBladeParent = Utils.GetChildNamed(_TRSaberMain, "blade_parent");
				_TRBlade2Parent = Utils.GetChildNamed(_TRSaberMain, "blade2_parent");
				_TRBladeTypes = Utils.GetChildNamed(_TRBladeParent, "blade_types");
				_TRBladeTypes2 = Utils.GetChildNamed(_TRBlade2Parent, "blade_types");
				_TRHiltParent = Utils.GetChildNamed(_TRSaberMain, "hilt_parent");
				
				_TRLoopSnd = Utils.GetChildNamed(_TRSaberMain, "Sound-loop");
				_TROnSnd = Utils.GetChildNamed(_TRSaberMain, "Sound-switch-on");
				_TROffSnd = Utils.GetChildNamed(_TRSaberMain, "Sound-switch-off");

				_ASLoopSnd = _TRLoopSnd.GetComponent<AudioSource>();
				_ASOnSnd = _TROnSnd.GetComponent<AudioSource>();
				_ASOffSnd = _TROffSnd.GetComponent<AudioSource>();
				
				_ANIMSaber = _TRSaberMain.GetComponent<Animation>();
				
				// Reading blade types for the choices
				_bladesChoices = new List<string>();
				foreach (Transform trTypes in _TRBladeTypes)
				{
					_bladesChoices.Add(trTypes.name);
				}
				
				UpdateBladeTypesChoices(SaberOpt_BladeA, _bladesChoices);
				UpdateBladeTypesChoices(SaberOpt_BladeB, _bladesChoices);
				
				// Grabbing content
				_hiltsChoices = new List<string>();
				_onSndChoices = new List<string>();
				_offSndChoices = new List<string>();
				_loopSndChoices = new List<string>();
				foreach (string bundleFileItem in bundleRequest.assetBundle.GetAllAssetNames())
				{
					if (bundleFileItem.Contains(HILTS_PATH))
					{
						Match pathMatch = Regex.Match(bundleFileItem, @"assets/vamlightsabers/materials/hilts/([^/]*)");
						string hiltName = pathMatch.Groups[1].Value;
						if (!_hiltsChoices.Contains(hiltName))
						{
							_hiltsChoices.Add(hiltName);							
						}
					}

					if (bundleFileItem.Contains(HILTS_PREFABS_PATH))
					{
						Match pathMatch = Regex.Match(bundleFileItem, @"assets/vamlightsabers/prefabs/hilts/([^/]*)/([^_]+)_");
						string partName = pathMatch.Groups[1].Value;
						string hiltName = pathMatch.Groups[2].Value;
						if (!_hiltsGameobjects.ContainsKey(partName))
						{
							List<GameObject> tmpListGO = new List<GameObject>();
							tmpListGO.Add(bundleRequest.assetBundle.LoadAsset<GameObject>(bundleFileItem));
							_hiltsGameobjects.Add(partName, tmpListGO);
						}
						else
						{
							_hiltsGameobjects[partName].Add(bundleRequest.assetBundle.LoadAsset<GameObject>(bundleFileItem));
						}
					}
					
					if (bundleFileItem.Contains(ON_SOUNDS_PATH))
					{
						Match pathMatch = Regex.Match(bundleFileItem, @"assets/vamlightsabers/sounds/on/([^/\.]*)");
						string sndName = pathMatch.Groups[1].Value;
						if (!_onSndChoices.Contains(sndName))
						{
							_onSndChoices.Add(sndName);
							_onSndSounds.Add(bundleRequest.assetBundle.LoadAsset<AudioClip>(bundleFileItem));
						}
					}
					
					if (bundleFileItem.Contains(OFF_SOUNDS_PATH))
					{
						Match pathMatch = Regex.Match(bundleFileItem, @"assets/vamlightsabers/sounds/off/([^/\.]*)");
						string sndName = pathMatch.Groups[1].Value;
						if (!_offSndChoices.Contains(sndName))
						{
							_offSndChoices.Add(sndName);
							_offSndSounds.Add(bundleRequest.assetBundle.LoadAsset<AudioClip>(bundleFileItem));
						}
					}
					
					if (bundleFileItem.Contains(LOOPS_SOUNDS_PATH))
					{
						Match pathMatch = Regex.Match(bundleFileItem, @"assets/vamlightsabers/sounds/loops/([^/\.]*)");
						string sndName = pathMatch.Groups[1].Value;
						if (!_loopSndChoices.Contains(sndName))
						{
							_loopSndChoices.Add(sndName);
							_loopSndSounds.Add(bundleRequest.assetBundle.LoadAsset<AudioClip>(bundleFileItem));
						}
					}
				}

				// Updating the parts selectors
				UpdatePartsChoices(SaberOpt_EmitterA, _hiltsChoices);
				UpdatePartsChoices(SaberOpt_HandleA, _hiltsChoices);
				UpdatePartsChoices(SaberOpt_SwitchA, _hiltsChoices);
				UpdatePartsChoices(SaberOpt_PowerCellA, _hiltsChoices);

				UpdatePartsChoices(SaberOpt_EmitterB, _hiltsChoices);
				UpdatePartsChoices(SaberOpt_HandleB, _hiltsChoices);
				UpdatePartsChoices(SaberOpt_SwitchB, _hiltsChoices);
				UpdatePartsChoices(SaberOpt_PowerCellB, _hiltsChoices);
				
				// Updating sounds selectors
				SaberOpt_OnSound.choices = _onSndChoices;
				SaberOpt_OffSound.choices = _offSndChoices;
				SaberOpt_LoopSound.choices = _loopSndChoices;
				
				_isBundleLoaded = true;
			}
			catch(Exception e)
			{
				SuperController.LogError(EXCEPTION_PREFIX + e);
			}
		}

		protected void OnLightsaberSwitchOn()
		{
			LightsaberSwitchOn();
		}
		
		protected void OnLightsaberSwitchOff()
		{
			LightsaberSwitchOff();
		}
		
		// **************************
		// LOCAL TOOLS
		// **************************
		private static void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private static void logError( string debugText ) {
			SuperController.LogError( debugText );
		}
		
		protected void ToggleUIDynamicGrp(List<UIDynamic> uiGrp,bool enabled)
		{
			foreach( UIDynamic uidElem in uiGrp )
			{
				ToggleUIDynamic(enabled, uidElem);
			}
		}
		
		protected void ToggleUIDynamic(bool enabled, UIDynamic target)
		{
			if (target == null) return;
			
			LayoutElement tmpLE = target.GetComponent<LayoutElement>();

			if (tmpLE != null)
			{
				if (enabled == true)
				{

					tmpLE.transform.localScale = new Vector3(1, 1, 1);
					if( !ignoreLayoutUI.Contains(target) ) tmpLE.ignoreLayout = false;
				}
				else
				{
					tmpLE.transform.localScale = new Vector3(0, 0, 0);
					tmpLE.ignoreLayout = true;
				}
			}
		}
		
		private void SetupButtonWithLayout(UIDynamicButton target, float newSize)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
		}
		
		private void SetupButtonWithoutLayout(UIDynamicButton target, float newSize, Vector2 newPosition)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
			tmpLE.ignoreLayout = true;
			tmpRectT.anchoredPosition = newPosition;
			ignoreLayoutUI.Add((UIDynamic) target);
		}

		// **************************
		// Time to cleanup !
		// **************************	
		void OnDisable() {

		}

		void OnDestroy()
		{
			// Cleaning GameObjects and textures
			Destroy(VAMLSGO);
			
			// Unloading assetbundle
			AssetLoader.DoneWithAssetBundleFromFile(PLUGIN_PATH + SABER_BUNDLE_PATH);
		}
    }
}